/*	AudioManager.java
 *	Author: Jimmy Yih, Mark Crompton 2011
 *	
 *	Core part of the application.  Does all the graphics, audio conversions,
 *	audio analysis, etc.
 */

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.geom.Line2D;
import java.awt.image.BufferedImage;
import java.awt.image.ColorConvertOp;
import java.awt.image.RescaleOp;
import java.awt.image.WritableRaster;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.util.Vector;

import javax.sound.sampled.AudioFileFormat;
import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.DataLine;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.SourceDataLine;
import javax.sound.sampled.TargetDataLine;
import javax.sound.sampled.UnsupportedAudioFileException;
import javax.swing.JPanel;

import Spectrogram.Spectrogram;

import com.mathworks.toolbox.javabuilder.MWClassID;
import com.mathworks.toolbox.javabuilder.MWComplexity;
import com.mathworks.toolbox.javabuilder.MWException;
import com.mathworks.toolbox.javabuilder.MWNumericArray;


public class AudioManager {

	private static boolean recording = false;
	private static boolean play = false;
	private static boolean pause = false;
	private static boolean playLines = false;
	private static Spectrogram s = null;
	public static byte[] output = null;

	// Params: byte[] data - Recorded or loaded audio data.
	// 		   AudioFormat format - Format of the recorded or loaded audio data.
	// Return: AudioInputStream - The audio stream formed from the recorded or loaded data for playback.
	// Notes: This is a more complete set of data for the recorded or loaded audio data.
	public static AudioInputStream getAudioInputStream(byte[] data, AudioFormat format){
		if(data == null || format == null)
			return null;
		ByteArrayInputStream bais = new ByteArrayInputStream(data);
		return new AudioInputStream(bais, format, data.length / format.getFrameSize());
	}

	// Params: AudioInputStream audioInputStream - The stream with the data to be extracted.
	// Return: byte[] - the audio data in a raw format.
	// Notes: This is for use in analyzing recorded or loaded audio data.
	public static byte[] getAudioBytes(AudioInputStream audioInputStream) throws IOException{
		byte[] data = new byte[(int)(audioInputStream.getFrameLength() * audioInputStream.getFormat().getFrameSize())];
		audioInputStream.read(data);
		return data;
	}

	// Params: AudioInputStream audioInputStream - The recorded or loaded data.
	//		   AudioFormat format - The format for the SourceDataLine.
	// Return: The SourceDataLine given the audio stream and format.
	// Notes: This method may take up to 1 second to complete.
	public static SourceDataLine getSourceDataLine(AudioInputStream audioInputStream) throws IOException, LineUnavailableException, UnsupportedAudioFileException{
		if(audioInputStream.markSupported())
			audioInputStream.reset();
		DataLine.Info info = new DataLine.Info(SourceDataLine.class, audioInputStream.getFormat());
		if (!AudioSystem.isLineSupported(info))
			return null;
		return (SourceDataLine) AudioSystem.getLine(info);
	}

	// Params: AudioFormat format - the format of the requested data line.
	// Return: The unopened TargetDataLine of the specified format, null if the line is not supported.
	// Notes: This method may take up to 1 second to complete.
	public static TargetDataLine getTargetDataLine(AudioFormat format) throws LineUnavailableException{
		DataLine.Info info = new DataLine.Info(TargetDataLine.class, format);
		boolean lines = AudioSystem.isLineSupported(info);
		if(!lines)
			return null;
		return (TargetDataLine) AudioSystem.getLine(info);
	}

	// Params: TargetDataLine line - The line that the data streams from.
	// Return: The byte data that was recorded.
	// Notes: This should be called in a separate thread from the event dispatching thread.
	public static byte[] startRecording(TargetDataLine line) throws LineUnavailableException, IOException{
		recording = true;
		ByteArrayOutputStream out = new ByteArrayOutputStream();

		int frameSizeInBytes = line.getFormat().getFrameSize();
		int bufferLengthInFrames = line.getBufferSize() / 8;
		int bufferLengthInBytes = bufferLengthInFrames * frameSizeInBytes;
		byte[] data = new byte[bufferLengthInBytes];
		int numBytesRead;

		line.open();
		line.start();
		while(recording){
			numBytesRead = line.read(data, 0, bufferLengthInBytes);
			out.write(data, 0, numBytesRead);
			output = out.toByteArray();
		}

		line.stop();
		line.close();

		out.flush();
		out.close();
		output = null;

		return out.toByteArray();
	}

	// Params: None
	// Return: None
	// Notes: Triggers the startRecording method to stop recording.  This may be called in any thread.
	public static void stopRecording(){
		recording = false;
	}

	// Params: AudioInputStream playbackInputStream - The audio stream to start playing.
	// Return: Will return false if the line is either invalid or unsupported.
	// Notes: It will not play if this method is running in another stream.
	public static boolean startPlayback(AudioInputStream playbackInputStream) throws LineUnavailableException, IOException, UnsupportedAudioFileException{
		if(play || playbackInputStream == null)
			return false;
		play = true;
		SourceDataLine line = getSourceDataLine(playbackInputStream);

		if(line == null)
			return false;

		line.open(line.getFormat(), 16384);

		int frameSizeInBytes = line.getFormat().getFrameSize();
		int bufferLengthInFrames = line.getBufferSize() / 8;
		int bufferLengthInBytes = bufferLengthInFrames * frameSizeInBytes;
		byte[] data = new byte[bufferLengthInBytes];
		int numBytesRead = 0;

		line.start();
		playing:
			while (play) {
				if ((numBytesRead = playbackInputStream.read(data)) == -1)
					break;
				int numBytesRemaining = numBytesRead;
				while (numBytesRemaining > 0 ){
					if(!play)
						break playing;
					if(pause) {
						line.stop();
						continue;
					}
					else
						line.start();
					numBytesRemaining -= line.write(data, 0, numBytesRemaining);
				}
			}

		if(play)
			line.drain();
		line.stop();
		line.close();
		play = false;
		return true;
	}

	public static class WaveformPanel extends JPanel{
		private static final long serialVersionUID = -7719666073063067151L;
		private Vector<Line2D> lines;

		public WaveformPanel(){
			super();
			lines = null;
			setBackground(Color.BLACK);
		}

		public WaveformPanel(Vector<Line2D> lines){
			this();
			this.lines = lines;
		}

		public void setLines(Vector<Line2D> lines){
			this.lines = lines;
			repaint();
		}

		public void paint(Graphics g){
			if(lines == null)
				super.paint(g);
			else{
				Dimension d = getSize();
				int w = d.width;
				int h = d.height;
				Graphics2D g2 = (Graphics2D) g;
				g2.setBackground(getBackground());
				g2.clearRect(0, 0, w, h);

				g2.setColor(Color.WHITE);
				for (int i = 1; i < lines.size(); i++) {
					g2.draw(lines.get(i));
				}
			}
		}

		// Draws the vertical line on the waveform panel
		public void VertLine(int duration) {
			Graphics g = this.getGraphics();
			Graphics2D g2 = (Graphics2D) g;
			int w = getSize().width;
			int h = getSize().height;
			playLines = true;
			long time1 = System.currentTimeMillis();
			long pauseTime = 0;
			long pauseDifference = 0;
			int prev = -1;
			while(playLines) {
				if (pause) {
					if (pauseTime == 0)
						pauseTime = System.currentTimeMillis();
					continue;
				}
				if (pauseTime != 0) {
					pauseDifference = (System.currentTimeMillis() - pauseTime);
					pauseTime = 0;
				}

				long time2 = System.currentTimeMillis();
				int current = (int) (time2 - time1 - pauseDifference);
				if (current >= duration) {
					playLines = false;
					break;
				}
				int i = (w * current) / duration;

				if (prev > 0) {
					g2.setColor(Color.BLACK);
					g2.drawLine(prev, 0, prev, h);
					g2.setColor(Color.WHITE);
					g2.draw(lines.get(prev));
				}
				else if (prev == 0) {
					g2.setColor(Color.BLACK);
					g2.drawLine(prev, 0, prev, h);
				}

				g2.setColor(Color.RED);
				g2.drawLine(i, 0, i, h);
				prev = i;

				try {
					Thread.sleep(100);
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
			g2.setColor(Color.BLACK);
			g2.drawLine(prev, 0, prev, h);
		}
	}

	// Creates the lines for the waveform panel
	public static Vector<Line2D> createWaveform(AudioInputStream audioInputStream, Dimension d) throws IOException{
		byte[] audioBytes = getAudioBytes(audioInputStream);

		Vector<Line2D> lines = new Vector<Line2D>();
		AudioFormat format = audioInputStream.getFormat();

		int w = d.width;
		int h = d.height;
		int[] audioData = null;
		if (format.getSampleSizeInBits() == 16) {
			int nlengthInSamples = audioBytes.length / 2;
			audioData = new int[nlengthInSamples];
			if (format.isBigEndian()) {
				for (int i = 0; i < nlengthInSamples; i++) {
					/* First byte is MSB (high order) */
					int MSB = (int) audioBytes[2*i];
					/* Second byte is LSB (low order) */
					int LSB = (int) audioBytes[2*i+1];
					audioData[i] = MSB << 8 | (255 & LSB);
				}
			} else {
				for (int i = 0; i < nlengthInSamples; i++) {
					/* First byte is LSB (low order) */
					int LSB = (int) audioBytes[2*i];
					/* Second byte is MSB (high order) */
					int MSB = (int) audioBytes[2*i+1];
					audioData[i] = MSB << 8 | (255 & LSB);
				}
			}
		} else if (format.getSampleSizeInBits() == 8) {
			int nlengthInSamples = audioBytes.length;
			audioData = new int[nlengthInSamples];
			if (format.getEncoding().toString().startsWith("PCM_SIGN")) {
				for (int i = 0; i < audioBytes.length; i++) {
					audioData[i] = audioBytes[i];
				}
			} else {
				for (int i = 0; i < audioBytes.length; i++) {
					audioData[i] = audioBytes[i] - 128;
				}
			}
		}

		int frames_per_pixel = audioBytes.length / format.getFrameSize()/w;
		byte my_byte = 0;
		double y_last = 0;
		int numChannels = format.getChannels();
		
		for (double x = 0; x < w && audioData != null; x++) {
			int idx = (int) (frames_per_pixel * numChannels * x);
			if (format.getSampleSizeInBits() == 8) {
				my_byte = (byte) audioData[idx];
			} else {
				my_byte = (byte) (128 * audioData[idx] / 32768 );
			}
			double y_new = (double) (h * (128 - my_byte) / 256);
			lines.add(new Line2D.Double(x, y_last, x, y_new));
			y_last = y_new;
		}

		return lines;
	}

	// Creates the data for the Spectrogram using Matlab packages
	public static double[][] createSpectrogram(AudioInputStream audioInputStream) throws IOException, MWException{
		double[] audioData = null;
		AudioFormat format = audioInputStream.getFormat();
		byte[] data = AudioManager.getAudioBytes(audioInputStream);
		s = new Spectrogram();

		if (format.getSampleSizeInBits() == 16) {
			int nlengthInSamples = data.length / 2;
			audioData = new double[nlengthInSamples];
			if (format.isBigEndian()) {
				for (int i = 0; i < nlengthInSamples; i++) {
					/* First byte is MSB (high order) */
					int MSB = (int) data[2*i];
					/* Second byte is LSB (low order) */
					int LSB = (int) data[2*i+1];
					audioData[i] = MSB << 8 | (255 & LSB);
					audioData[i] /= 32768;
				}
			} else {
				for (int i = 0; i < nlengthInSamples; i++) {
					/* First byte is LSB (low order) */
					int LSB = (int) data[2*i];
					/* Second byte is MSB (high order) */
					int MSB = (int) data[2*i+1];
					audioData[i] = MSB << 8 | (255 & LSB);
					audioData[i] /= 32768;
				}
			}
		} else if (format.getSampleSizeInBits() == 8) {
			int nlengthInSamples = data.length;
			audioData = new double[nlengthInSamples];
			if (format.getEncoding().toString().startsWith("PCM_SIGN")) {
				for (int i = 0; i < data.length; i++) {
					int dat = data[i];
					audioData[i] = dat / 128.0;
				}
			} else {
				for (int i = 0; i < data.length; i++) {
					int dat = data[i];
					audioData[i] = (dat<0?(dat+128):(dat-128)) / 128.0;
				}
			}
		}
		int[] dims = {audioData.length, 1};
		MWNumericArray arr = MWNumericArray.newInstance(dims, MWClassID.DOUBLE, MWComplexity.REAL);
		arr.setData(audioData);
		Object[] result = s.specgramN(1, arr, audioInputStream.getFormat().getSampleRate());
		MWNumericArray output = (MWNumericArray)result[0];
		double[][] outputM = (double[][]) output.toDoubleArray();
		double[][] flipOutput = new double[outputM[0].length][outputM.length];
		for(int i = 0; i < outputM.length; i++){
			for(int j = 0; j < outputM[0].length; j++){
				flipOutput[j][outputM.length-1-i] = outputM[i][j];
			}
		}
		return flipOutput;
	}
	
	// Uses the spectrogram data and creates an image from it (simulates Matlab code)
	public static BufferedImage getSpectrogramImage(double[][] spectrogram){
		double max = Double.MIN_VALUE;
		double min = Double.MAX_VALUE;
		for(int i = 0; i < spectrogram.length; i++){
			for(int j = 0; j < spectrogram[0].length; j++){
				max = Math.max(max, spectrogram[i][j]);
				min = Math.min(min, spectrogram[i][j]);
			}
		}
		double[][] imageData = new double[spectrogram.length][spectrogram[0].length];
		BufferedImage image = new BufferedImage(imageData.length, imageData[0].length, BufferedImage.TYPE_BYTE_INDEXED);
		WritableRaster wr = (WritableRaster) image.getData();

		for(int i = 0; i < spectrogram.length; i++){
			for(int j = 0; j < spectrogram[0].length; j++){
				imageData[i][j] = spectrogram[i][j] * 1000;
				double[] pixel = {imageData[i][j]};
				wr.setPixel(i, j, pixel);
			}
		}
		image.setData(wr);

		BufferedImage gray = new BufferedImage(imageData.length, imageData[0].length, BufferedImage.TYPE_BYTE_GRAY);
		ColorConvertOp op = new ColorConvertOp(image.getColorModel().getColorSpace(),gray.getColorModel().getColorSpace(),null);
		op.filter(image,gray);

		RescaleOp rOp = new RescaleOp(-1.0f, 255f, null);
		BufferedImage invert = rOp.filter(gray, null);

		return invert;
	}


	// Params: String filename - The name of the file to load.
	// Return: The AudioInputStream of the resultant read file.
	// Notes: Loads any wav, au, or aif/aiff file into an AudioInputStream.
	public static AudioInputStream load(String filename) throws UnsupportedAudioFileException, IOException{
		return AudioSystem.getAudioInputStream(new File(filename));
	}

	// Params: AudioInputStream audioInputStream - The audio data to be saved.
	//         String filename - the name of the file to be saved to.
	// Return: None.
	// Notes: This should probably only be used to save recorded audio data.
	public static void saveToWave(AudioInputStream audioInputStream, String filename) throws IOException{
		AudioSystem.write(audioInputStream, AudioFileFormat.Type.WAVE, new File(filename));
	}

	// Params: None
	// Return: None
	// Notes: Pauses playback if playing.
	public static void pausePlayback(){
		if(play)
			pause = true;
	}

	// Params: None
	// Return: None
	// Notes: Resumes playback if playing.
	public static void resumePlayback(){
		if(pause)
			pause = false;
	}

	// Params: None
	// Return: None
	// Notes: Stops playback if playing.
	public static void stopPlayback(){
		play = false;
		playLines = false;
		pause = false;
	}

	// Params: None
	// Return: Returns true if the AudioManager is recording audio.
	// Notes: Persists through threads.
	public static boolean isRecording(){
		return recording;
	}

	// Params: None
	// Return: Returns true if the AudioManager is playing audio.
	// Notes: Persists through threads.
	public static boolean isPlaying(){
		return play;
	}

	// Params: None
	// Return: Returns true if the AudioManager has paused playback.
	// Notes: Persists through threads.
	public static boolean isPaused(){
		return pause;
	}
}
